from django.db import migrations

payroll_rights = [202001, 202002, 202004]
imis_administrator_system = 64


def add_rights(apps, schema_editor):
    role = apps.get_model('core', 'role').objects.get(is_system=imis_administrator_system)
    for right_id in payroll_rights:
        if not apps.get_model('core', 'roleright').objects.filter(validity_to__isnull=True, role=role,
                                                                  right_id=right_id).exists():
            _add_right_for_role(apps, role, right_id)


def _add_right_for_role(apps, role, right_id):
    apps.get_model('core', 'roleright').objects.create(role=role, right_id=right_id, audit_user_id=1)


def remove_rights(apps, schema_editor):
    apps.get_model('core', 'roleright').objects.filter(
        role__is_system=imis_administrator_system,
        right_id__in=payroll_rights,
        validity_to__isnull=True
    ).delete()


class Migration(migrations.Migration):
    dependencies = [
        ('payroll', '0003_historicalpayroll_historicalpayrollbill_payroll_payrollbill'),
    ]

    operations = [
        migrations.RunPython(add_rights, remove_rights),
    ]



from django.urls import path

from payroll.views import send_callback_to_openimis, CSVReconciliationAPIView

urlpatterns = [
    path('send_callback_to_openimis/', send_callback_to_openimis),
    path('csv_reconciliation/', CSVReconciliationAPIView.as_view()),
]



import json
import uuid
from datetime import datetime, timedelta
from core.models import MutationLog

from graphene import JSONString, Schema
from graphene.test import Client
from django.test import TestCase
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q

from contribution_plan.models import PaymentPlan
from individual.models import Individual
from individual.tests.data import service_add_individual_payload
from invoice.models import Bill
from payment_cycle.models import PaymentCycle
from payroll.models import Payroll, PayrollBill, PayrollStatus
from payroll.tests.data import gql_payroll_create, gql_payroll_query, gql_payroll_delete, \
    gql_payroll_create_no_json_ext
from payroll.tests.helpers import PaymentPointHelper
from core.test_helpers import LogInHelper
from payroll.schema import Query, Mutation
from social_protection.models import BenefitPlan, Beneficiary, BeneficiaryStatus
from social_protection.tests.data import service_add_payload
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase, BaseTestContext


class PayrollGQLTestCase(openIMISGraphQLTestCase):

    user = None
    user_unauthorized = None
    gql_client = None
    gql_context = None
    gql_context_unauthorized = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user = LogInHelper().get_or_create_user_api(username='username_authorized', roles=[7])
        cls.user_unauthorized = LogInHelper().get_or_create_user_api(username='username_unauthorized', roles=[1])
        gql_schema = Schema(
            query=Query,
            mutation=Mutation
        )
        cls.gql_client = Client(gql_schema)
        cls.gql_context = BaseTestContext(cls.user)
        cls.gql_context_unauthorized = BaseTestContext(cls.user_unauthorized)
        cls.name = "TestCreatePayroll"
        cls.status = PayrollStatus.PENDING_APPROVAL
        cls.payment_method = "TestPaymentMethod"
        cls.date_valid_from, cls.date_valid_to = cls.__get_start_and_end_of_current_month()
        cls.payment_point = PaymentPointHelper().get_or_create_payment_point_api()
        cls.benefit_plan = cls.__create_benefit_plan()
        cls.individual = cls.__create_individual()
        cls.subject_type = ContentType.objects.get_for_model(Beneficiary)
        cls.payment_plan = cls.__create_payment_plan(cls.benefit_plan)
        cls.payment_cycle = cls.__create_payment_cycle()
        cls.beneficiary = cls.__create_beneficiary()
        cls.bill = cls.__create_bill()
        cls.json_ext_able_bodied_false = """{"advanced_criteria": [{"custom_filter_condition": "able_bodied__boolean=False"}]}"""
        cls.json_ext_able_bodied_true = """{"advanced_criteria": [{"custom_filter_condition": "able_bodied__boolean=True"}]}"""
        cls.includedUnpaid = False

    def setup(self):
        payroll = self.payroll_from_db()
        if payroll:
            self.delete_payroll_and_check_bill(payroll)
        temp_payroll = self.payroll_from_db(f"{self.name}-tmp")
        if temp_payroll:
            self.delete_payroll_and_check_bill(temp_payroll)

    def payroll_from_db(self, name=None):
        return Payroll.objects.filter(
            name=(name or self.name),
            payment_plan_id=self.payment_plan.id,
            payment_cycle_id=self.payment_cycle.id,
            payment_point_id=self.payment_point.id,
            payment_method=self.payment_method,
            status=self.status,
            date_valid_from=self.date_valid_from,
            date_valid_to=self.date_valid_to,
            is_deleted=False,
        ).first()

    def test_query(self):
        output = self.gql_client.execute(gql_payroll_query, context=self.gql_context.get_request())
        result = output.get('data', {}).get('payroll', {})
        self.assertTrue(result)

    def test_query_unauthorized(self):
        output = self.gql_client.execute(gql_payroll_query, context=self.gql_context_unauthorized.get_request())
        error = next(iter(output.get('errors', [])), {}).get('message', None)
        self.assertTrue(error)

    def create_payroll(self, name, json_ext):
        variables = {
            "name": name,
            "paymentCycleId": str(self.payment_cycle.id),
            "paymentPlanId": str(self.payment_plan.id),
            "paymentPointId": str(self.payment_point.id),
            "paymentMethod": self.payment_method,
            "status": self.status,
            "dateValidFrom": self.date_valid_from,
            "dateValidTo": self.date_valid_to,
            "jsonExt": json_ext,
            "paymentPointId": str(self.payment_point.id),
            "clientMutationId": str(uuid.uuid4())
        }
        output = self.gql_client.execute(gql_payroll_create, context=self.gql_context.get_request(), variable_values=variables)
        self.assertEqual(output.get('errors'), None)

        return self.payroll_from_db(name)

    def create_payroll_no_json_ext(self, name):
        payload = gql_payroll_create_no_json_ext % (
            name,
            self.payment_cycle.id,
            self.payment_plan.id,
            self.payment_point.id,
            self.payment_method,
            self.status,
            self.date_valid_from,
            self.date_valid_to,
        )
        output = self.gql_client.execute(payload, context=self.gql_context.get_request())
        self.assertEqual(output.get('errors'), None)

        return self.payroll_from_db(name)

    def delete_payroll_and_check_bill(self, payroll):
        # payroll_bill = PayrollBill.objects.filter(
        #     bill=self.bill, payroll=payroll.first())
        # self.assertEqual(payroll_bill.count(), 1)
        # payroll_bill.delete()
        payroll.delete(username='username_authorized')
        # self.assertEqual(PayrollBill.objects.all().count(), 0)
        # FIXME
        # self.assertIsNone(payroll)

    # def test_create_fail_due_to_lack_of_bills_for_given_criteria(self):
    #     payroll = self.create_payroll(self.name, self.json_ext_able_bodied_false)
    #     self.assertFalse(payroll.exists())

    def test_create_no_advanced_criteria(self):
        payroll = self.create_payroll_no_json_ext(self.name)
        self.assertIsNotNone(payroll)
        self.delete_payroll_and_check_bill(payroll)

    def test_create_full(self):
        payroll = self.create_payroll(self.name, self.json_ext_able_bodied_true)
        self.assertIsNotNone(payroll)
        self.delete_payroll_and_check_bill(payroll)

    def test_create_fail_due_to_empty_name(self):
        payroll = self.create_payroll("", self.json_ext_able_bodied_true)
        self.assertIsNone(payroll)

    # def test_create_fail_due_to_one_bill_assigment(self):
    #     tmp_name = f"{self.name}-tmp"
    #     payroll_tmp = self.create_payroll(tmp_name, self.json_ext_able_bodied_true)
    #     self.assertTrue(payroll_tmp.exists())

    #     payroll = self.create_payroll(self.name, self.json_ext_able_bodied_true)
    #     self.assertFalse(payroll.exists())

    #     self.delete_payroll_and_check_bill(payroll_tmp)

    def test_create_unauthorized(self):
        variables = {
            "name": self.name,
            "paymentCycleId": str(self.payment_cycle.id),
            "paymentPlanId": str(self.payment_plan.id),
            "paymentMethod": self.payment_method,
            "status": PayrollStatus.PENDING_APPROVAL,
            "dateValidFrom": self.date_valid_from,
            "dateValidTo": self.date_valid_to,
            "jsonExt": self.json_ext_able_bodied_false,
            "clientMutationId": str(uuid.uuid4())
        }

        output = self.gql_client.execute(
            gql_payroll_create, context=self.gql_context_unauthorized.get_request(), variable_values=variables)
        self.assertFalse(
            Payroll.objects.filter(
                name=self.name,
                payment_plan_id=self.benefit_plan.id,
                payment_cycle_id=self.payment_cycle.id,
                payment_point_id=self.payment_point.id,
                payment_method=self.payment_method,
                status=self.status,
                date_valid_from=self.date_valid_from,
                date_valid_to=self.date_valid_to,
                json_ext=self.json_ext_able_bodied_false,
                is_deleted=False,
            ).exists()
        )

    def test_delete(self):
        payroll = Payroll(name=self.name,
                          payment_plan_id=self.payment_plan.id,
                          payment_point_id=self.payment_point.id,
                          payment_cycle_id=self.payment_cycle.id,
                          payment_method=self.payment_method,
                          status=self.status,
                          date_valid_from=self.date_valid_from,
                          date_valid_to=self.date_valid_to,
                          json_ext=json.loads(self.json_ext_able_bodied_false),
                          )
        payroll.save(username=self.user.username)
        # payroll_bill = PayrollBill(payroll=payroll, bill=self.bill)
        # payroll_bill.save(username=self.user.username)
        payload = gql_payroll_delete % json.dumps([str(payroll.id)])
        output = self.gql_client.execute(payload, context=self.gql_context.get_request())
        self.assertEqual(output.get('errors'), None)
        # FIXME self.assertTrue(Payroll.objects.filter(id=payroll.id, is_deleted=True).exists())
        # FIXME 
        # self.assertEqual(PayrollBill.objects.filter(payroll=payroll, bill=self.bill).count(), 0)

    def test_delete_unauthorized(self):
        payroll = Payroll(name=self.name,
                          payment_plan_id=self.payment_plan.id,
                          payment_cycle_id=self.payment_cycle.id,
                          payment_point_id=self.payment_point.id,
                          payment_method=self.payment_method,
                          status=self.status,
                          date_valid_from=self.date_valid_from,
                          date_valid_to=self.date_valid_to,
                          json_ext=json.loads(self.json_ext_able_bodied_true),
                          )
        payroll.save(username=self.user.username)
        payroll_bill = PayrollBill(payroll=payroll, bill=self.bill)
        payroll_bill.save(username=self.user.username)
        payload = gql_payroll_delete % json.dumps([str(payroll.id)])
        output = self.gql_client.execute(payload, context=self.gql_context_unauthorized.get_request())
        # FIXME look for delete task instead
        # self.assertTrue(Payroll.objects.filter(id=payroll.id, is_deleted=False).exists())
        # self.assertEqual(PayrollBill.objects.filter(payroll=payroll, bill=self.bill).count(), 1)
        # payroll_bill.delete(username=self.user.username)
        # self.assertTrue(PayrollBill.objects.filter(payroll=payroll, bill=self.bill, is_deleted=True))

    @classmethod
    def __create_benefit_plan(cls):
        object_data = {
            **service_add_payload
        }

        benefit_plan = BenefitPlan(**object_data)
        benefit_plan.save(username=cls.user.username)

        return benefit_plan

    @classmethod
    def __create_payment_plan(cls, benefit_plan):
        object_data = {
            'is_deleted': False,
            'code': "PP-E",
            'name': "Example Payment Plan",
            'benefit_plan': benefit_plan,
            'periodicity': 1,
            'calculation': "32d96b58-898a-460a-b357-5fd4b95cd87c",
            'json_ext': {
                'calculation_rule': {
                    'fixed_batch': 2,
                    'limit_per_single_transaction': 100
                } 
            },
        }

        payment_plan = PaymentPlan(**object_data)
        payment_plan.save(username=cls.user.username)
        return payment_plan

    @classmethod
    def __create_payment_cycle(cls):
        pc = PaymentCycle(
            start_date='2023-02-01', 
            end_date='2023-03-01', 
            type=ContentType.objects.get_for_model(BenefitPlan),
            code=str(datetime.now()))
        pc.save(username=cls.user.username)
        return pc

    @classmethod
    def __create_individual(cls):
        object_data = {
            **service_add_individual_payload
        }

        individual = Individual(**object_data)
        individual.save(username=cls.user.username)

        return individual

    @classmethod
    def __create_beneficiary(cls):
        object_data = {
            "individual": cls.individual,
            "benefit_plan": cls.benefit_plan,
            "json_ext": {"able_bodied": True},
            "status": BeneficiaryStatus.ACTIVE
        }
        beneficiary = Beneficiary(**object_data)
        beneficiary.save(username=cls.user.username)
        return beneficiary

    @classmethod
    def __create_bill(cls):
        object_data = {
            "subject_type": cls.subject_type,
            "subject_id": cls.beneficiary.id,
            "status": Bill.Status.VALIDATED,
            "code": str(datetime.now())
        }
        bill = Bill(**object_data)
        bill.save(username=cls.user.username)
        return bill

    @classmethod
    def __get_start_and_end_of_current_month(cls):
        today = datetime.today()
        # Set date_valid_from to the beginning of the current month
        date_valid_from = today.replace(day=1).strftime("%Y-%m-%d")

        # Calculate the last day of the current month
        next_month = today.replace(day=28) + timedelta(days=4)  # Adding 4 days to ensure we move to the next month
        date_valid_to = (next_month - timedelta(days=next_month.day)).strftime("%Y-%m-%d")

        return date_valid_from, date_valid_to



from django.apps import apps
from django.conf import settings

is_unit_test_env = getattr(settings, 'IS_UNIT_TEST_ENV', False)

# Check if the 'opensearch_reports' app is in INSTALLED_APPS
if 'opensearch_reports' in apps.app_configs and not is_unit_test_env:
    from opensearch_reports.service import BaseSyncDocument
    from django_opensearch_dsl import fields as opensearch_fields
    from django_opensearch_dsl.registries import registry
    from payroll.models import (
        Payroll,
        PayrollBenefitConsumption,
        BenefitConsumption,
        BenefitAttachment
    )
    from payment_cycle.models import PaymentCycle
    from contribution_plan.models import PaymentPlan
    from individual.models import Individual
    from invoice.models import Bill

    @registry.register_document
    class PayrollDocument(BaseSyncDocument):
        DASHBOARD_NAME = 'Payment'

        name = opensearch_fields.KeywordField()
        status = opensearch_fields.KeywordField()
        payment_method = opensearch_fields.KeywordField()
        date_created = opensearch_fields.DateField()
        payment_plan = opensearch_fields.ObjectField(properties={
            'code': opensearch_fields.KeywordField(),
            'name': opensearch_fields.KeywordField(),
        })
        payment_cycle = opensearch_fields.ObjectField(properties={
            'code': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'start_date': opensearch_fields.DateField(),
            'end_date': opensearch_fields.DateField(),
        })

        class Index:
            name = 'payroll'
            settings = {
                'number_of_shards': 1,
                'number_of_replicas': 0
            }
            auto_refresh = True

        class Django:
            model = Payroll
            fields = [
                'id'
            ]
            related_models = [PaymentPlan, PaymentCycle]
            queryset_pagination = 5000

        def get_instances_from_related(self, related_instance):
            if isinstance(related_instance, PaymentPlan):
                return Payroll.objects.filter(payment_plan=related_instance)
            elif isinstance(related_instance, PaymentCycle):
                return Payroll.objects.filter(payment_cycle=related_instance)


    @registry.register_document
    class BenefitConsumptionDocument(BaseSyncDocument):
        DASHBOARD_NAME = 'Payment'

        photo = opensearch_fields.KeywordField()
        code = opensearch_fields.KeywordField()
        date_due = opensearch_fields.DateField()
        date_created = opensearch_fields.DateField()
        receipt = opensearch_fields.KeywordField()
        amount = opensearch_fields.KeywordField()
        type = opensearch_fields.KeywordField()
        status = opensearch_fields.KeywordField()
        json_ext = opensearch_fields.ObjectField()
        individual = opensearch_fields.ObjectField(properties={
            'first_name': opensearch_fields.KeywordField(),
            'last_name': opensearch_fields.KeywordField(),
            'dob': opensearch_fields.DateField(),
        })

        class Index:
            name = 'benefit_consumption'
            settings = {
                'number_of_shards': 1,
                'number_of_replicas': 0
            }
            auto_refresh = True

        class Django:
            model = BenefitConsumption
            related_models = [Individual]
            fields = [
                'id'
            ]
            queryset_pagination = 5000

        def get_instances_from_related(self, related_instance):
            if isinstance(related_instance, Individual):
                return BenefitConsumption.objects.filter(individual=related_instance)


    @registry.register_document
    class PayrollBenefitConsumptionDocument(BaseSyncDocument):
        DASHBOARD_NAME = 'Payment'

        payroll = opensearch_fields.ObjectField(properties={
            'name': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'payment_method': opensearch_fields.KeywordField(),
            'date_created': opensearch_fields.DateField(),
            'payment_plan': opensearch_fields.ObjectField(properties={
                'code': opensearch_fields.KeywordField(),
                'name': opensearch_fields.KeywordField(),
            }),
            'payment_cycle': opensearch_fields.ObjectField(properties={
                'code': opensearch_fields.KeywordField(),
                'status': opensearch_fields.KeywordField(),
                'start_date': opensearch_fields.DateField(),
                'end_date': opensearch_fields.DateField(),
            })
        })
        benefit = opensearch_fields.ObjectField(properties={
            'code': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'type': opensearch_fields.KeywordField(),
            'receipt': opensearch_fields.KeywordField(),
            'amount': opensearch_fields.KeywordField(),
            'photo': opensearch_fields.KeywordField(),
            'date_due': opensearch_fields.DateField(),
            'individual': opensearch_fields.ObjectField(properties={
              'first_name': opensearch_fields.KeywordField(),
              'last_name': opensearch_fields.KeywordField(),
              'dob': opensearch_fields.DateField(),
            })
        })

        class Index:
            name = 'payroll_benefit_consumption'
            settings = {
                'number_of_shards': 1,
                'number_of_replicas': 0
            }
            auto_refresh = True

        class Django:
            model = PayrollBenefitConsumption
            related_models = [Payroll, BenefitConsumption]
            fields = [
                'id'
            ]
            queryset_pagination = 5000

        def get_instances_from_related(self, related_instance):
            if isinstance(related_instance, Payroll):
                return PayrollBenefitConsumption.objects.filter(payroll=related_instance)
            elif isinstance(related_instance, BenefitConsumption):
                return PayrollBenefitConsumption.objects.filter(benefit=related_instance)


    @registry.register_document
    class BenefitAttachmentDocument(BaseSyncDocument):
        DASHBOARD_NAME = 'Invoice'

        bill = opensearch_fields.ObjectField(properties={
            'code': opensearch_fields.KeywordField(),
            'code_ext': opensearch_fields.KeywordField(),
            'code_tp': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'currency_code': opensearch_fields.KeywordField(),
            'note': opensearch_fields.KeywordField(),
            'terms': opensearch_fields.KeywordField(),
            'date_created': opensearch_fields.DateField(),
            'date_due': opensearch_fields.DateField(),
            'date_payed': opensearch_fields.DateField(),
            'amount_total': opensearch_fields.KeywordField(),
        })
        benefit = opensearch_fields.ObjectField(properties={
            'code': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'type': opensearch_fields.KeywordField(),
            'receipt': opensearch_fields.KeywordField(),
            'amount': opensearch_fields.KeywordField(),
            'photo': opensearch_fields.KeywordField(),
            'date_due': opensearch_fields.DateField(),
            'individual': opensearch_fields.ObjectField(properties={
                'first_name': opensearch_fields.KeywordField(),
                'last_name': opensearch_fields.KeywordField(),
                'dob': opensearch_fields.DateField(),
            })
        })
        payroll = opensearch_fields.NestedField(properties={
            'name': opensearch_fields.KeywordField(),
            'status': opensearch_fields.KeywordField(),
            'payment_method': opensearch_fields.KeywordField(),
            'date_created': opensearch_fields.DateField(),
            'payment_plan': opensearch_fields.ObjectField(properties={
                'code': opensearch_fields.KeywordField(),
                'name': opensearch_fields.KeywordField(),
            }),
            'payment_cycle': opensearch_fields.ObjectField(properties={
                'code': opensearch_fields.KeywordField(),
                'status': opensearch_fields.KeywordField(),
                'start_date': opensearch_fields.DateField(),
                'end_date': opensearch_fields.DateField(),
            })
        })

        class Index:
            name = 'benefit_attachment'
            settings = {
                'number_of_shards': 1,
                'number_of_replicas': 0
            }
            auto_refresh = True

        class Django:
            model = BenefitAttachment
            related_models = [Payroll, Bill, BenefitConsumption]
            fields = [
                'id'
            ]
            queryset_pagination = 5000

        def get_instances_from_related(self, related_instance):
            if isinstance(related_instance, Payroll):
                return BenefitAttachment.objects.filter(
                    benefit__payrollbenefitconsumption__payroll=related_instance
                )
            elif isinstance(related_instance, Bill):
                return BenefitAttachment.objects.filter(bill=related_instance)
            elif isinstance(related_instance, BenefitConsumption):
                return BenefitAttachment.objects.filter(benefit=related_instance)



import graphene
from django.db.models import Sum, Q
from graphene_django import DjangoObjectType

from core import prefix_filterset, ExtendedConnection
from core.gql_queries import UserGQLType
from core.utils import DefaultStorageFileHandler
from invoice.gql.gql_types.bill_types import BillGQLType
from location.gql_queries import LocationGQLType
from individual.gql_queries import IndividualGQLType
from payroll.models import PaymentPoint, Payroll, BenefitConsumption, \
    PayrollBenefitConsumption, BenefitAttachment, CsvReconciliationUpload
from contribution_plan.gql import PaymentPlanGQLType
from payment_cycle.gql_queries import PaymentCycleGQLType
from social_protection.models import BenefitPlan


class PaymentPointGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')

    class Meta:
        model = PaymentPoint
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "name": ["iexact", "istartswith", "icontains"],
            **prefix_filterset("location__", LocationGQLType._meta.filter_fields),
            **prefix_filterset("ppm__", UserGQLType._meta.filter_fields),

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection


class BenefitAttachmentGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')

    class Meta:
        model = BenefitAttachment
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            **prefix_filterset("bill__", BillGQLType._meta.filter_fields),

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_from": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_to": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection


class BenefitConsumptionGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')
    benefit_attachment = graphene.List(BenefitAttachmentGQLType)

    class Meta:
        model = BenefitConsumption
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "photo": ["iexact", "istartswith", "icontains"],
            "code": ["iexact", "istartswith", "icontains"],
            "status": ["exact", "startswith", "icontains", "contains"],
            "receipt": ["exact", "startswith", "icontains"],
            "type": ["exact", "startswith", "icontains"],
            "amount": ["exact", "lt", "lte", "gt", "gte"],
            "date_due": ["exact", "lt", "lte", "gt", "gte"],
            **prefix_filterset("individual__", IndividualGQLType._meta.filter_fields),

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_from": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_to": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection

    def resolve_benefit_attachment(self, info):
        return BenefitAttachment.objects.filter(
            benefit_id=self.id,
            is_deleted=False
        )


class PayrollGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')
    benefit_consumption = graphene.List(BenefitConsumptionGQLType)
    benefit_plan_name_code = graphene.String()

    class Meta:
        model = Payroll
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "name": ["iexact", "istartswith", "icontains"],
            "status": ["exact", "startswith", "contains"],
            "payment_method": ["exact", "startswith", "contains"],
            **prefix_filterset("payment_point__", PaymentPointGQLType._meta.filter_fields),
            **prefix_filterset("payment_plan__", PaymentPlanGQLType._meta.filter_fields),
            **prefix_filterset("payment_cycle__", PaymentCycleGQLType._meta.filter_fields),

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_from": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_to": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection

    def resolve_benefit_consumption(self, info):
        return BenefitConsumption.objects.filter(payrollbenefitconsumption__payroll__id=self.id,
                                                 is_deleted=False,
                                                 payrollbenefitconsumption__is_deleted=False)

    def resolve_benefit_plan_name_code(self, info):
        benefit_plan = BenefitPlan.objects.get(id=self.payment_plan.benefit_plan.id, is_deleted=False)
        return f"{benefit_plan.code} - {benefit_plan.name}"


class PaymentMethodGQLType(graphene.ObjectType):
    name = graphene.String()


class PaymentGatewayConfigGQLType(graphene.ObjectType):
    base_url = graphene.String()
    api_key = graphene.String()
    timeout = graphene.Int()


class PaymentMethodListGQLType(graphene.ObjectType):
    payment_methods = graphene.List(PaymentMethodGQLType)


class BenefitAttachmentListGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')

    class Meta:
        model = BenefitAttachment
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            **prefix_filterset("bill__", BillGQLType._meta.filter_fields),
            **prefix_filterset("benefit__", BenefitConsumptionGQLType._meta.filter_fields),

            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_from": ["exact", "lt", "lte", "gt", "gte"],
            "date_valid_to": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection


class CsvReconciliationUploadGQLType(DjangoObjectType):
    uuid = graphene.String(source='uuid')

    class Meta:
        model = CsvReconciliationUpload
        interfaces = (graphene.relay.Node,)

        filter_fields = {
            "id": ["exact"],
            "file_name": ["exact", "iexact", "istartswith", "icontains"],
            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "status": ["exact", "iexact", "istartswith", "icontains"],
            "is_deleted": ["exact"],
            "version": ["exact"],
            **prefix_filterset("payroll__", PayrollGQLType._meta.filter_fields),
        }
        connection_class = ExtendedConnection


class PayrollBenefitConsumptionGQLType(DjangoObjectType):

    class Meta:
        model = PayrollBenefitConsumption
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            **prefix_filterset("payroll__", PayrollGQLType._meta.filter_fields),
            **prefix_filterset("benefit__", BenefitConsumptionGQLType._meta.filter_fields),
            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }
        connection_class = ExtendedConnection


class BenefitsSummaryGQLType(graphene.ObjectType):
    total_amount_received = graphene.String()
    total_amount_due = graphene.String()



# Generated by Django 3.2.21 on 2024-02-12 13:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('payroll', '0017_csvreconciliationupload_historicalcsvreconciliationupload'),
    ]

    operations = [
        migrations.AlterField(
            model_name='benefitconsumption',
            name='receipt',
            field=models.CharField(blank=True, db_column='Receipt', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalbenefitconsumption',
            name='receipt',
            field=models.CharField(blank=True, db_column='Receipt', max_length=255, null=True),
        ),
    ]



# Generated by Django 3.2.19 on 2023-08-11 07:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('payroll', '0006_auto_20230809_1057'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalpayroll',
            name='payment_method',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='payroll',
            name='payment_method',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



import abc


class StrategyOfPaymentInterface(object,  metaclass=abc.ABCMeta):

    @classmethod
    def initialize_payment_gateway(cls):
        pass

    @classmethod
    def accept_payroll(cls, payroll, user, **kwargs):
        pass

    @classmethod
    def make_payment_for_payroll(cls, payroll, user, **kwargs):
        pass

    @classmethod
    def reject_payroll(cls, payroll, user, **kwargs):
        from payroll.models import PayrollStatus
        cls.change_status_of_payroll(payroll, PayrollStatus.REJECTED, user)
        cls.remove_benefits_from_rejected_payroll(payroll)

    @classmethod
    def reject_approved_payroll(cls, payroll, user):
        from django.contrib.contenttypes.models import ContentType
        from core.services.utils.serviceUtils import model_representation
        from payroll.models import (
            BenefitConsumption,
            BenefitConsumptionStatus,
            PayrollStatus
        )
        from invoice.models import (
            DetailPaymentInvoice,
            PaymentInvoice,
            Bill
        )
        from payroll.services import PayrollService

        benefit_data = BenefitConsumption.objects.filter(
            payrollbenefitconsumption__payroll=payroll,
            status=BenefitConsumptionStatus.RECONCILED,
            is_deleted=False
        )
        benefit_data_related = list(benefit_data.values_list('id', 'benefitattachment__bill'))
        if len(benefit_data_related) > 0:
            benefits, related_bills = zip(*benefit_data_related)
            bill_content_type = ContentType.objects.get_for_model(Bill)
            detail_payment_invoices = DetailPaymentInvoice.objects.filter(
                subject_type=bill_content_type,
                subject_id__in=related_bills
            )
            payment_invoice_ids = list(detail_payment_invoices.values_list('payment_id', flat=True))
            detail_payment_invoices.delete()
            PaymentInvoice.objects.filter(id__in=payment_invoice_ids).delete()

        for benefit in benefit_data:
            benefit.receipt = None
            benefit.status = BenefitConsumptionStatus.ACCEPTED
            benefit.save(username=user.username)
        cls.change_status_of_payroll(payroll, PayrollStatus.PENDING_APPROVAL, user)
        PayrollService(user).create_accept_payroll_task(payroll.id, model_representation(payroll))

    @classmethod
    def acknowledge_of_reponse_view(cls, payroll, response_from_gateway, user, rejected_bills):
        pass

    @classmethod
    def reconcile_payroll(cls, payroll, user):
        pass

    @classmethod
    def change_status_of_payroll(cls, payroll, status, user):
        payroll.status = status
        payroll.save(username=user.login_name)

    @classmethod
    def remove_benefits_from_rejected_payroll(cls, payroll):
        from payroll.models import (
            BenefitAttachment,
            BenefitConsumption,
            PayrollBenefitConsumption,
        )
        from invoice.models import (
            Bill,
            BillItem
        )

        benefit_data = BenefitConsumption.objects.filter(
            payrollbenefitconsumption__payroll=payroll,
            is_deleted=False
        ).values_list('id', 'benefitattachment__bill')

        if len(benefit_data) > 0:
            benefits, related_bills = zip(*benefit_data)

            BenefitAttachment.objects.filter(
                benefit_id__in=benefits
            ).delete()

            BillItem.objects.filter(
                bill__id__in=related_bills
            ).delete()

            Bill.objects.filter(
                id__in=related_bills
            ).delete()

            PayrollBenefitConsumption.objects.filter(payroll=payroll).delete()

            BenefitConsumption.objects.filter(
                id__in=benefits,
                is_deleted=False
            ).delete()

    @classmethod
    def remove_benefit_from_payroll(cls, benefit):
        from payroll.models import (
            BenefitAttachment,
            BenefitConsumption,
            PayrollBenefitConsumption
        )
        from invoice.models import (
            Bill,
            BillItem
        )

        benefit_data = BenefitConsumption.objects.filter(
            id=benefit.id,
            is_deleted=False
        ).values_list('id', 'benefitattachment__bill')

        if len(benefit_data) > 0:
            benefits, related_bills = zip(*benefit_data)

            BenefitAttachment.objects.filter(
                benefit_id__in=benefits
            ).delete()

            BillItem.objects.filter(
                bill__id__in=related_bills
            ).delete()

            Bill.objects.filter(
                id__in=related_bills
            ).delete()

            PayrollBenefitConsumption.objects.filter(benefit=benefit).delete()

            BenefitConsumption.objects.filter(
                id__in=benefits,
                is_deleted=False
            ).delete()



# Generated by Django 3.2.19 on 2023-08-09 10:57

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('payroll', '0005_change_i_user_to_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalpayroll',
            name='status',
            field=models.CharField(choices=[('CREATED', 'CREATED'), ('ONGOING', 'ONGOING'), ('AWAITING_FOR_RECONCILIATION', 'AWAITING_FOR_RECONCILIATION'), ('RECONCILIATED', 'RECONCILIATED')], default='CREATED', max_length=100),
        ),
        migrations.AddField(
            model_name='payroll',
            name='status',
            field=models.CharField(choices=[('CREATED', 'CREATED'), ('ONGOING', 'ONGOING'), ('AWAITING_FOR_RECONCILIATION', 'AWAITING_FOR_RECONCILIATION'), ('RECONCILIATED', 'RECONCILIATED')], default='CREATED', max_length=100),
        ),
        migrations.AlterField(
            model_name='paymentpoint',
            name='ppm',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='core.user'),
        ),
    ]



import logging

from django.db.models import Q, Sum
from django.db import transaction

from core.signals import register_service_signal
from payroll.strategies.strategy_of_payments_interface import StrategyOfPaymentInterface
from payroll.utils import CodeGenerator

logger = logging.getLogger(__name__)


class StrategyOnlinePayment(StrategyOfPaymentInterface):
    WORKFLOW_NAME = "payment-adaptor"
    WORKFLOW_GROUP = "openimis-coremis-payment-adaptor"
    PAYMENT_GATEWAY = None

    @classmethod
    def initialize_payment_gateway(cls):
        from payroll.payment_gateway import PaymentGatewayConfig
        gateway_config = PaymentGatewayConfig()
        payment_gateway_connector_class = gateway_config.get_payment_gateway_connector()
        cls.PAYMENT_GATEWAY = payment_gateway_connector_class()

    @classmethod
    def accept_payroll(cls, payroll, user, **kwargs):
        cls._process_accepted_payroll(payroll, user, **kwargs)

    @classmethod
    def make_payment_for_payroll(cls, payroll, user, **kwargs):
        cls._send_payment_data_to_gateway(payroll, user)

    @classmethod
    def acknowledge_of_reponse_view(cls, payroll, response_from_gateway, user, rejected_bills):
        # save response coming from payment gateway in json_ext
        cls._save_payroll_data(payroll, user, response_from_gateway)

    @classmethod
    @transaction.atomic
    def reconcile_payroll(cls, payroll, user):
        from payroll.tasks import send_request_to_reconcile
        send_request_to_reconcile.delay(payroll.id, user.id)

    @classmethod
    def get_benefits_attached_to_payroll(cls, payroll, status):
        from payroll.models import BenefitConsumption
        filters = Q(
            payrollbenefitconsumption__payroll_id=payroll.id,
            is_deleted=False,
            status=status,
            payrollbenefitconsumption__is_deleted=False,
            payrollbenefitconsumption__payroll__is_deleted=False,
        )
        benefits = BenefitConsumption.objects.filter(filters)
        return benefits

    @classmethod
    def approve_for_payment_benefit_consumption(cls, benefits, user):
        from payroll.models import BenefitConsumptionStatus
        for benefit in benefits:
            try:
                benefit.status = BenefitConsumptionStatus.APPROVE_FOR_PAYMENT
                benefit.save(username=user.login_name)
            except Exception as e:
                logger.debug(f"Failed to approve benefit consumption {benefit.code}: {str(e)}")

    @classmethod
    def reconcile_benefit_consumption(cls, benefits, user):
        from payroll.models import BenefitConsumptionStatus
        from payroll.apps import PayrollConfig
        from invoice.models import Bill
        for benefit in benefits:
            try:
                receipt = CodeGenerator.generate_unique_code(
                    'payroll',
                    'BenefitConsumption',
                    'receipt',
                    PayrollConfig.receipt_length,
                )
                benefit.receipt = receipt
                benefit.status = BenefitConsumptionStatus.RECONCILED
                benefit.save(username=user.login_name)
                bill = Bill.objects.filter(
                    benefitattachment__benefit=benefit,
                    is_deleted=False
                ).first()
                if bill:
                    cls._create_bill_payment_for_paid_bill(benefit, bill, user)
            except Exception as e:
                logger.debug(f"Failed to approve benefit consumption {benefit.code}: {str(e)}")

    @classmethod
    def _create_bill_payment_for_paid_bill(cls, benefit, bill, user):
        from core import datetime
        from django.contrib.contenttypes.models import ContentType
        from invoice.models import Bill, DetailPaymentInvoice, PaymentInvoice
        from invoice.services import PaymentInvoiceService
        current_date = datetime.date.today()
        bill.status = Bill.Status.RECONCILIATED
        bill.date_payed = current_date
        bill.save(username=user.login_name)
        # Create a BillPayment object for the 'Paid' bill
        bill_payment = {
            "code_tp": bill.code_tp,
            "code_ext": bill.code_ext,
            "code_receipt": bill.code,
            "label": bill.terms,
            'reconciliation_status': PaymentInvoice.ReconciliationStatus.RECONCILIATED,
            "fees": 0.0,
            "amount_received": bill.amount_total,
            "date_payment": current_date,
            'payment_origin': "online payment",
            'payer_ref': 'payment reference',
            'payer_name': 'payer name',
            "json_ext": {}
        }

        bill_payment_details = {
            'subject_type': ContentType.objects.get_for_model(bill),
            'subject': bill,
            'status': DetailPaymentInvoice.DetailPaymentStatus.ACCEPTED,
            'fees': 0.0,
            'amount': bill.amount_total,
            'reconcilation_id': benefit.receipt,
            'reconcilation_date': current_date,
        }
        bill_payment_details = DetailPaymentInvoice(**bill_payment_details)
        payment_service = PaymentInvoiceService(user)
        payment_service.create_with_detail(bill_payment, bill_payment_details)

    @classmethod
    def _get_payroll_bills_amount(cls, payroll):
        from payroll.models import Payroll
        payroll_with_benefit_sum = Payroll.objects.filter(id=payroll.id).annotate(
            total_benefit_amount=Sum('payrollbenefitconsumption__benefit__amount')
        ).first()
        return payroll_with_benefit_sum.total_benefit_amount

    @classmethod
    def _get_benefits_to_string(cls, benefits):
        benefits_uuids = [str(benefit.id) for benefit in benefits]
        benefits_uuids_string = ",".join(benefits_uuids)
        return benefits_uuids_string

    @classmethod
    def _send_payment_data_to_gateway(cls, payroll, user):
        from payroll.models import BenefitConsumptionStatus
        benefits = cls.get_benefits_attached_to_payroll(payroll, BenefitConsumptionStatus.ACCEPTED)
        payment_gateway_connector = cls.PAYMENT_GATEWAY
        benefits_to_approve = []
        for benefit in benefits:
            if payment_gateway_connector.send_payment(benefit.code, benefit.amount):
                benefits_to_approve.append(benefit)
            else:
                # Handle the case where a benefit payment is rejected
                logger.info(f"Payment for benefit ({benefit.code}) was rejected.")
        if benefits_to_approve:
            cls.approve_for_payment_benefit_consumption(benefits_to_approve, user)

    @classmethod
    def _process_accepted_payroll(cls, payroll, user, **kwargs):
        from payroll.models import PayrollStatus
        cls.change_status_of_payroll(payroll, PayrollStatus.APPROVE_FOR_PAYMENT, user)

    @classmethod
    def _save_payroll_data(cls, payroll, user, response_from_gateway):
        json_ext = payroll.json_ext if payroll.json_ext else {}
        json_ext['response_from_gateway'] = response_from_gateway
        payroll.json_ext = json_ext
        payroll.save(username=user.username)
        cls._create_payroll_reconcilation_task(payroll, user)

    @classmethod
    @register_service_signal('online_payments.create_task')
    def _create_payroll_reconcilation_task(cls, payroll, user):
        from payroll.apps import PayrollConfig
        from tasks_management.services import TaskService
        from tasks_management.apps import TasksManagementConfig
        from tasks_management.models import Task
        TaskService(user).create({
            'source': 'payroll_reconciliation',
            'entity': payroll,
            'status': Task.Status.RECEIVED,
            'executor_action_event': TasksManagementConfig.default_executor_event,
            'business_event': PayrollConfig.payroll_reconciliation_event,
        })



import importlib

from payroll.apps import PayrollConfig


class PaymentGatewayConfig:
    def __init__(self):
        self.gateway_base_url = PayrollConfig.gateway_base_url
        self.endpoint_payment = PayrollConfig.endpoint_payment
        self.endpoint_reconciliation = PayrollConfig.endpoint_reconciliation
        self.api_key = PayrollConfig.payment_gateway_api_key
        self.basic_auth_username = PayrollConfig.payment_gateway_basic_auth_username
        self.basic_auth_password = PayrollConfig.payment_gateway_basic_auth_password
        self.timeout = PayrollConfig.payment_gateway_timeout
        self.auth_type = PayrollConfig.payment_gateway_auth_type

    def get_headers(self):
        if self.auth_type == 'token':
            return {
                'Authorization': f'Bearer {self.api_key}',
                'Content-Type': 'application/json',
            }
        elif self.auth_type == 'basic':
            import base64
            auth_str = f"{self.basic_auth_username}:{self.basic_auth_password}"
            auth_bytes = auth_str.encode('utf-8')
            auth_base64 = base64.b64encode(auth_bytes).decode('utf-8')
            return {
                'Authorization': f'Basic {auth_base64}',
                'Content-Type': 'application/json',
            }
        else:
            return {
                'Content-Type': 'application/json',
            }

    def get_payment_gateway_connector(self):
        module_name, class_name = PayrollConfig.payment_gateway_class.rsplit('.', 1)
        module = importlib.import_module(module_name)
        return getattr(module, class_name)

    def get_payment_endpoint(self):
        return self.endpoint_payment

    def get_reconciliation_endpoint(self):
        return self.endpoint_reconciliation



# Generated by Django 4.2.20 on 2025-04-04 10:02

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("payroll", "0021_auto_20240715_0855"),
    ]

    operations = [
        migrations.AlterField(
            model_name="csvreconciliationupload",
            name="user_created",
            field=models.ForeignKey(
                db_column="UserCreatedUUID",
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="%(class)s_user_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="csvreconciliationupload",
            name="user_updated",
            field=models.ForeignKey(
                db_column="UserUpdatedUUID",
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="%(class)s_user_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Generated by Django 3.2.21 on 2023-10-06 07:55

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('payroll', '0008_historicalpaymentadaptorhistory_paymentadaptorhistory'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payroll',
            name='payment_point',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='payroll.paymentpoint'),
        ),
    ]



from payroll.strategies.strategy_of_payments_interface import StrategyOfPaymentInterface


class StrategyOfflinePayment(StrategyOfPaymentInterface):

    @classmethod
    def accept_payroll(cls, payroll, user, **kwargs):
        from payroll.models import PayrollStatus
        cls.change_status_of_payroll(payroll, PayrollStatus.APPROVE_FOR_PAYMENT, user)

    @classmethod
    def reconcile_payroll(cls, payroll, user):
        from payroll.models import PayrollStatus
        cls.change_status_of_payroll(payroll, PayrollStatus.RECONCILED, user)



# Generated by Django 3.2.21 on 2024-02-07 09:07

import core.models
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_mutationlog_autogenerated_code'),
        ('payroll', '0012_auto_20240202_1619'),
    ]

    operations = [
        migrations.CreateModel(
            name='PayrollMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='payroll', to='core.mutationlog')),
                ('payroll', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='payroll.payroll')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
    ]



from payroll.strategies.strategy_of_payments_interface import StrategyOfPaymentInterface


class StrategyMobilePayment(StrategyOfPaymentInterface):

    @classmethod
    def accept_payroll(cls, payroll, user, **kwargs):
        pass



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _

from core.validation import BaseModelValidation
from payroll.models import PaymentPoint, Payroll, PayrollBill, BenefitConsumption


class PaymentPointValidation(BaseModelValidation):
    OBJECT_TYPE = PaymentPoint

    @classmethod
    def validate_create(cls, user, **data):
        super().validate_create(user, **data)

    @classmethod
    def validate_update(cls, user, **data):
        super().validate_update(user, **data)

    @classmethod
    def validate_delete(cls, user, **data):
        super().validate_delete(user, **data)


class PayrollValidation(BaseModelValidation):
    OBJECT_TYPE = Payroll

    @classmethod
    def validate_create(cls, user, **data):
        errors = validate_payroll(data)
        if errors:
            raise ValidationError(errors)
        super().validate_create(user, **data)


class BenefitConsumptionValidation(BaseModelValidation):
    OBJECT_TYPE = BenefitConsumption

    @classmethod
    def validate_create(cls, user, **data):
        super().validate_create(user, **data)


def validate_payroll(data):
    return [
        *are_bills_in_data(data),
        *validate_one_payroll_per_bill(data),
        *validate_payroll_unique_name(data, uuid=None),
        *validate_not_empty_field(data, 'name')
    ]


def are_bills_in_data(data):
    bills = data.get('bills', None)
    if not bills:
        return [{"message": _("payroll.validation.payroll.no_bills_for_filter_criteria")}]
    return []


def validate_one_payroll_per_bill(data):
    bills = data.get('bills', [])
    query = PayrollBill.objects.filter(bill__in=bills, is_deleted=False)
    if query.exists():
        payroll_bill_ids = list(query.values_list('id', flat=True))
        return [{"message": _("payroll.validation.payroll.bill_already_assigned") % {
            "bill_ids": payroll_bill_ids,
        }}]
    return []


def validate_payroll_unique_name(data, uuid=None):
    name = data.get("name")
    instance = Payroll.objects.filter(name=name, is_deleted=False).first()
    if instance and instance.uuid != uuid:
        return [{"message": _("payroll.validation.payroll.name_exists" % {
            'name': name
        })}]
    return []


def validate_not_empty_field(data, field):
    string = data.get('name')
    if not string:
        return [{"message": _("payroll.validation.field_empty") % {
            'field': field
        }}]
    return []



import logging
from typing import List

from payroll.payments_registry.registry_point import PaymentsMethodRegistryPoint
from payroll.strategies import StrategyOfPaymentInterface

logger = logging.getLogger(__name__)


class PaymentMethodStorage:

    @classmethod
    def get_all_available_payment_methods(cls) -> List[StrategyOfPaymentInterface]:
        return PaymentsMethodRegistryPoint.REGISTERED_PAYMENT_METHODS

    @classmethod
    def get_chosen_payment_method(cls, payment_method_name: str) -> StrategyOfPaymentInterface:
        chosen_method = None
        for method_info in PaymentsMethodRegistryPoint.REGISTERED_PAYMENT_METHODS:
            if method_info['name'] == payment_method_name:
                chosen_method = method_info['class_reference']
        return chosen_method



import os

from django.apps import AppConfig

from core.custom_filters import CustomFilterRegistryPoint
from payroll.payments_registry import PaymentsMethodRegistryPoint

MODULE_NAME = 'payroll'

DEFAULT_CONFIG = {
    "gql_payment_point_search_perms": ["201001"],
    "gql_payment_point_create_perms": ["201002"],
    "gql_payment_point_update_perms": ["201003"],
    "gql_payment_point_delete_perms": ["201004"],
    "gql_payroll_search_perms": ["202001"],
    "gql_payroll_create_perms": ["202002"],
    "gql_payroll_delete_perms": ["202004"],
    "gql_csv_reconciliation_search_perms": ["206001"],
    "gql_csv_reconciliation_create_perms": ["206002"],
    "payroll_accept_event": "payroll.accept_payroll",
    "payroll_reconciliation_event": "payroll.payroll_reconciliation",
    "payroll_reject_event": "payroll.payroll_reject",
    "csv_reconciliation_field_mapping": {
        'payrollbenefitconsumption__payroll__name': 'Payroll Name',
        'payrollbenefitconsumption__payroll__status': 'Payroll Status',
        'individual__first_name': 'First Name',
        'individual__last_name': 'Last Name',
        'individual__dob': 'Date of Birth',
        'code': 'Code',
        'status': 'Status',
        'amount': 'Amount',
        'type': 'Type',
        'receipt': 'Receipt',
    },
    "csv_reconciliation_status_column": "Status",
    "csv_reconciliation_paid_extra_field": "Paid",
    "csv_reconciliation_receipt_column": "receipt",
    "csv_reconciliation_errors_column": "errors",
    "csv_reconciliation_code_column": "code",
    "csv_reconciliation_paid_yes": "Yes",
    "csv_reconciliation_paid_no": "No",
    "payroll_delete_event": "payroll.payroll_delete",
    "benefit_delete_event": "payroll.benefit_delete",

    "gateway_base_url": "http://41.175.18.170:8070/api/mobile/v1/",
    "endpoint_payment": "mock/payment",
    "endpoint_reconciliation": "mock/reconciliation",
    "payment_gateway_api_key": os.getenv('PAYMENT_GATEWAY_API_KEY'),
    "payment_gateway_basic_auth_username": os.getenv('PAYMENT_GATEWAY_BASIC_AUTH_USERNAME'),
    "payment_gateway_basic_auth_password": os.getenv('PAYMENT_GATEWAY_BASIC_AUTH_PASSWORD'),
    "payment_gateway_timeout": 5,
    "payment_gateway_auth_type": "basic",  # can be 'token' or 'basic'
    "payment_gateway_class": "payroll.payment_gateway.MockedPaymentGatewayConnector",
    "receipt_length": 8
}


class PayrollConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = MODULE_NAME

    gql_payment_point_search_perms = None
    gql_payment_point_create_perms = None
    gql_payment_point_update_perms = None
    gql_payment_point_delete_perms = None
    gql_payroll_search_perms = None
    gql_payroll_create_perms = None
    gql_payroll_delete_perms = None
    gql_csv_reconciliation_search_perms = None
    gql_csv_reconciliation_create_perms = None
    payroll_accept_event = None
    payroll_reconciliation_event = None
    payroll_reject_event = None
    csv_reconciliation_field_mapping = None
    csv_reconciliation_status_column = None
    csv_reconciliation_paid_extra_field = None
    csv_reconciliation_receipt_column = None
    csv_reconciliation_errors_column = None
    csv_reconciliation_code_column = None
    csv_reconciliation_paid_yes = None
    csv_reconciliation_paid_no = None
    payroll_delete_event = None
    benefit_delete_event = None

    gateway_base_url = None
    endpoint_payment = None
    endpoint_reconciliation = None
    payment_gateway_api_key = None
    payment_gateway_basic_auth_username = None
    payment_gateway_basic_auth_password = None
    payment_gateway_timeout = None
    payment_gateway_auth_type = None
    payment_gateway_class = None
    receipt_length = None

    def ready(self):
        from core.models import ModuleConfiguration

        cfg = ModuleConfiguration.get_or_default(self.name, DEFAULT_CONFIG)
        self.__load_config(cfg)
        self.__register_filters_and_payment_methods()

    @classmethod
    def __load_config(cls, cfg):
        """
        Load all config fields that match current AppConfig class fields, all custom fields have to be loaded separately
        """
        for field in cfg:
            if hasattr(PayrollConfig, field):
                setattr(PayrollConfig, field, cfg[field])

    def __register_filters_and_payment_methods(cls):
        from social_protection.custom_filters import BenefitPlanCustomFilterWizard
        CustomFilterRegistryPoint.register_custom_filters(
            module_name=cls.name,
            custom_filter_class_list=[BenefitPlanCustomFilterWizard]
        )

        from payroll.strategies import (
            StrategyOfflinePayment,
            StrategyOnlinePayment
        )
        PaymentsMethodRegistryPoint.register_payment_method(
            payment_method_class_list=[
                StrategyOfflinePayment(),
                StrategyOnlinePayment(),
            ]
        )

    @staticmethod
    def get_payroll_payment_file_path(payroll_id, file_name=None):
        if file_name:
            return f"csv_reconciliation/payroll_{payroll_id}/{file_name}"
        return f"csv_reconciliation/payroll_{payroll_id}"



import graphene
from gettext import gettext as _
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from django.db import transaction

from core.gql.gql_mutations.base_mutation import BaseHistoryModelCreateMutationMixin, BaseMutation, \
    BaseHistoryModelUpdateMutationMixin, BaseHistoryModelDeleteMutationMixin
from core.schema import OpenIMISMutation
from payroll.apps import PayrollConfig
from payroll.models import PaymentPoint, Payroll, PayrollStatus, PayrollMutation
from payroll.services import PaymentPointService, PayrollService, BenefitConsumptionService


class CreatePaymentPointInputType(OpenIMISMutation.Input):
    name = graphene.String(required=True, max_length=255)
    location_id = graphene.Int(required=True)
    ppm_id = graphene.UUID(required=True)


class UpdatePaymentPointInputType(CreatePaymentPointInputType):
    id = graphene.UUID(required=True)


class DeletePaymentPointInputType(OpenIMISMutation.Input):
    ids = graphene.List(graphene.UUID, required=True)


class UpdatePaymentGatewayConfigInputType(OpenIMISMutation.Input):
    base_url = graphene.String(required=True, max_length=255)
    api_key = graphene.String(required=True, max_length=255)
    timeout = graphene.Int(required=True)


class CreatePayrollInput(OpenIMISMutation.Input):
    class PayrollStatusEnum(graphene.Enum):
        PENDING_APPROVAL = PayrollStatus.PENDING_APPROVAL
        APPROVE_FOR_PAYMENT = PayrollStatus.APPROVE_FOR_PAYMENT
        REJECTED = PayrollStatus.REJECTED
        RECONCILED = PayrollStatus.RECONCILED

    name = graphene.String(required=True, max_length=255)
    payment_plan_id = graphene.UUID(required=True)
    payment_point_id = graphene.UUID(required=False)
    payment_cycle_id = graphene.UUID(required=False)
    status = graphene.Field(PayrollStatusEnum, required=True)
    payment_method = graphene.String(required=True, max_length=255)
    from_failed_invoices_payroll_id = graphene.UUID(required=False)

    date_valid_from = graphene.Date(required=False)
    date_valid_to = graphene.Date(required=False)
    json_ext = graphene.JSONString(required=False)


class DeletePayrollInputType(DeletePaymentPointInputType):
    pass


class ClosePayrollInputType(DeletePaymentPointInputType):
    pass


class CreatePaymentPointMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreatePaymentPointMutation"
    _mutation_module = PayrollConfig.name
    _model = PaymentPoint

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                PayrollConfig.gql_payment_point_create_perms):
            raise ValidationError(_("mutation.authentication_required"))

    @classmethod
    def _mutate(cls, user, **data):
        data.pop('client_mutation_id', None)
        data.pop('client_mutation_label', None)

        service = PaymentPointService(user)
        service.create(data)

    class Input(CreatePaymentPointInputType):
        pass


class UpdatePaymentPointMutation(BaseHistoryModelUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdatePaymentPointMutation"
    _mutation_module = PayrollConfig.name
    _model = PaymentPoint

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                PayrollConfig.gql_payment_point_update_perms):
            raise ValidationError(_("mutation.authentication_required"))

    @classmethod
    def _mutate(cls, user, **data):
        data.pop('client_mutation_id', None)
        data.pop('client_mutation_label', None)

        service = PaymentPointService(user)
        service.update(data)

    class Input(UpdatePaymentPointInputType):
        pass


class DeletePaymentPointMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeletePaymentPointMutation"
    _mutation_module = PayrollConfig.name
    _model = PaymentPoint

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(
                PayrollConfig.gql_payment_point_delete_perms):
            raise ValidationError(_("mutation.authentication_required"))

    @classmethod
    def _mutate(cls, user, **data):
        data.pop('client_mutation_id', None)
        data.pop('client_mutation_label', None)

        service = PaymentPointService(user)

        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.delete({'id': id, 'user': user})

    class Input(DeletePaymentPointInputType):
        pass


class CreatePayrollMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "CreatePayrollMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.has_perms(
                PayrollConfig.gql_payroll_create_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.pop('client_mutation_id', None)
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = PayrollService(user)
        response = service.create(data)
        if client_mutation_id and response['success']:
            payroll_id = response['data']['id']
            payroll = Payroll.objects.get(id=payroll_id)
            PayrollMutation.object_mutated(
                user, client_mutation_id=client_mutation_id, payroll=payroll
            )
        if not response['success']:
            return response
        return None

    class Input(CreatePayrollInput):
        pass


class DeletePayrollMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeletePayrollMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.has_perms(
                PayrollConfig.gql_payroll_delete_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = PayrollService(user)
        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.delete({'id': id, 'user': user})

    class Input(DeletePayrollInputType):
        pass


class ClosePayrollMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "ClosePayrollMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.has_perms(
                PayrollConfig.gql_payroll_delete_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = PayrollService(user)
        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.close_payroll({'id': id})

    class Input(DeletePayrollInputType):
        pass


class MakePaymentForPayrollMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "MakePaymentForPayrollMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        super()._validate_mutation(user, **data)
        if not user.has_perms(PayrollConfig.gql_payroll_create_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = PayrollService(user)
        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.make_payment_for_payroll({'id': id})

    class Input(DeletePayrollInputType):
        pass


class RejectPayrollMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "RejectPayrollMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.has_perms(
                PayrollConfig.gql_payroll_delete_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = PayrollService(user)
        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.reject_approved_payroll({'id': id})

    class Input(DeletePayrollInputType):
        pass


class DeleteBenefitConsumptionMutation(BaseHistoryModelDeleteMutationMixin, BaseMutation):
    _mutation_class = "DeleteBenefitConsumptionMutation"
    _mutation_module = "payroll"
    _model = Payroll

    @classmethod
    def _validate_mutation(cls, user, **data):
        if type(user) is AnonymousUser or not user.has_perms(
                PayrollConfig.gql_payroll_delete_perms):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop('client_mutation_id')
        if "client_mutation_label" in data:
            data.pop('client_mutation_label')

        service = BenefitConsumptionService(user)
        ids = data.get('ids')
        if ids:
            with transaction.atomic():
                for id in ids:
                    service.delete({'id': id})

    class Input(DeletePayrollInputType):
        pass




import logging

from core.models import User
from core.service_signals import ServiceSignalBindType
from core.signals import bind_service_signal
from openIMIS.openimisapps import openimis_apps
from tasks_management.models import Task
from payroll.apps import PayrollConfig
from payroll.models import Payroll, BenefitConsumption, BenefitConsumptionStatus
from payroll.payments_registry import PaymentMethodStorage
from payroll.services import PayrollService
from payroll.strategies import StrategyOfPaymentInterface


logger = logging.getLogger(__name__)
imis_modules = openimis_apps()


def bind_service_signals():
    def on_task_complete_accept_payroll(**kwargs):
        def accept_payroll(payroll, strategy, user):
            if strategy:
                strategy.accept_payroll(payroll, user)

        def reject_payroll(payroll, strategy, user):
            if strategy:
                strategy.reject_payroll(payroll, user)

        try:
            result = kwargs.get('result', None)
            task = result['data']['task']
            user = User.objects.get(id=result['data']['user']['id'])
            if result \
                    and result['success'] \
                    and task['business_event'] == PayrollConfig.payroll_accept_event:
                task_status = task['status']
                payroll = Payroll.objects.get(id=task['entity_id'])
                strategy = PaymentMethodStorage.get_chosen_payment_method(payroll.payment_method)
                if task_status == Task.Status.COMPLETED:
                    accept_payroll(payroll, strategy, user)
                if task_status == Task.Status.FAILED:
                    reject_payroll(payroll, strategy, user)
        except Exception as exc:
            logger.error("Error while executing on_task_complete_accept_payroll", exc_info=exc)

    def on_task_complete_payroll_reconcilation(**kwargs):
        def reconcile_payroll(payroll, user):
            strategy = PaymentMethodStorage.get_chosen_payment_method(payroll.payment_method)
            if strategy:
                strategy.reconcile_payroll(payroll, user)
        try:
            result = kwargs.get('result', None)
            task = result['data']['task']
            user = User.objects.get(id=result['data']['user']['id'])
            if result \
                    and result['success'] \
                    and task['business_event'] == PayrollConfig.payroll_reconciliation_event:
                task_status = task['status']
                if task_status == Task.Status.COMPLETED:
                    payroll = Payroll.objects.get(id=task['entity_id'])
                    reconcile_payroll(payroll, user)
        except Exception as exc:
            logger.error("Error while executing on_task_complete_payroll_reconciliation", exc_info=exc)

    def on_task_complete_payroll_reject_approved_payroll(**kwargs):
        def reject_approved_payroll(payroll, user):
            strategy = PaymentMethodStorage.get_chosen_payment_method(payroll.payment_method)
            if strategy:
                strategy.reject_approved_payroll(payroll, user)
        try:
            result = kwargs.get('result', None)
            task = result['data']['task']
            user = User.objects.get(id=result['data']['user']['id'])
            if result \
                    and result['success'] \
                    and task['business_event'] == PayrollConfig.payroll_reject_event:
                task_status = task['status']
                if task_status == Task.Status.COMPLETED:
                    payroll = Payroll.objects.get(id=task['entity_id'])
                    reject_approved_payroll(payroll, user)
        except Exception as exc:
            logger.error("Error while executing on_task_complete_reject_approved_payroll", exc_info=exc)

    def on_task_delete_payroll(**kwargs):
        def delete_payroll(payroll, user):
            strategy = PaymentMethodStorage.get_chosen_payment_method(payroll.payment_method)
            if strategy:
                strategy.remove_benefits_from_rejected_payroll(payroll=payroll)
                PayrollService(user).delete_instance(payroll)
        try:
            result = kwargs.get('result', None)
            task = result['data']['task']
            user = User.objects.get(id=result['data']['user']['id'])
            if result \
                    and result['success'] \
                    and task['business_event'] == PayrollConfig.payroll_delete_event:
                task_status = task['status']
                if task_status == Task.Status.COMPLETED:
                    payroll = Payroll.objects.get(id=task['entity_id'])
                    delete_payroll(payroll, user)
        except Exception as exc:
            logger.error("Error while executing on_task_complete_delete_payroll", exc_info=exc)

    def on_task_delete_benefit(**kwargs):
        def delete_benefit(benefit, user):
            StrategyOfPaymentInterface.remove_benefit_from_payroll(benefit=benefit)
        try:
            result = kwargs.get('result', None)
            task = result['data']['task']
            user = User.objects.get(id=result['data']['user']['id'])
            if result \
                    and result['success'] \
                    and task['business_event'] == PayrollConfig.benefit_delete_event:
                task_status = task['status']
                if task_status == Task.Status.COMPLETED:
                    benefit = BenefitConsumption.objects.get(id=task['entity_id'])
                    delete_benefit(benefit, user)
                if task_status == Task.Status.FAILED:
                    benefit = BenefitConsumption.objects.get(id=task['entity_id'])
                    benefit.status = BenefitConsumptionStatus.ACCEPTED
                    benefit.save(username=user.username)
        except Exception as exc:
            logger.error("Error while executing on_task_complete_delete_benefit", exc_info=exc)

    bind_service_signal(
        'task_service.complete_task',
        on_task_complete_accept_payroll,
        bind_type=ServiceSignalBindType.AFTER
    )

    bind_service_signal(
        'task_service.complete_task',
        on_task_complete_payroll_reconcilation,
        bind_type=ServiceSignalBindType.AFTER
    )

    bind_service_signal(
        'task_service.complete_task',
        on_task_complete_payroll_reject_approved_payroll,
        bind_type=ServiceSignalBindType.AFTER
    )

    bind_service_signal(
        'task_service.complete_task',
        on_task_delete_payroll,
        bind_type=ServiceSignalBindType.AFTER
    )

    bind_service_signal(
        'task_service.complete_task',
        on_task_delete_benefit,
        bind_type=ServiceSignalBindType.AFTER
    )



from core.models import User
from core.services import create_or_update_interactive_user, create_or_update_core_user
from location.models import Location
from payroll.models import PaymentPoint
from payroll.services import PaymentPointService
from core.test_helpers import LogInHelper


class PaymentPointHelper:
    _TEST_DATA_PAYMENT_POINT = None
    def __init__(self):
        self._TEST_DATA_PAYMENT_POINT=        {
        "name": "TestPaymentPoint",
        "location": Location.objects.filter(validity_to__isnull=True, type='V').first()
    }

    def get_or_create_payment_point_api(self, **kwargs):
        payment_point = PaymentPoint.objects.filter(name={**self._TEST_DATA_PAYMENT_POINT, **kwargs}["name"]).first()
        if payment_point is None:
            payment_point = self.__create_payment_point(**kwargs)
        return payment_point

    def __create_payment_point(self, **kwargs):
        user = LogInHelper().get_or_create_user_api()
        name = {**self._TEST_DATA_PAYMENT_POINT, **kwargs}["name"],
        location = {**self._TEST_DATA_PAYMENT_POINT, **kwargs}["location"]
        payment_point = PaymentPoint(
            name=name,
            location=location,
            ppm=user
        )
        payment_point.save(username=user.username)
        return payment_point



from datetime import date


gql_payment_point_query = """
query q1 {
  paymentPoint {
    edges {
      node {
        id
      }
    }
  }
}
"""

gql_payment_point_filter = """
query q1 {
  paymentPoint(name_Iexact: "%s", location_Uuid: "%s", ppm_Uuid: "%s") {
    edges {
      node {
        id
      }
    }
  }
}
"""

gql_payment_point_create = """
mutation m1 {
  createPaymentPoint (input:{
    name: %s,
    locationId: %s,
    ppmId: "%s"
  }) {
    clientMutationId
  }
}
"""

gql_payment_point_update = """
mutation m1 {
  updatePaymentPoint (input:{
    id: %s
    name: %s,
    locationId: %s,
    ppmId: "%s"
  }) {
    clientMutationId
  }
}
"""

gql_payment_point_delete = """
mutation m1 {
  deletePaymentPoint (input:{
    ids: %s
  }) {
    clientMutationId
  }
}
"""

gql_payroll_query = """
query q2 {
  payroll {
    edges {
      node {
        id
      }
    }
  }
}
"""

gql_payroll_filter = """
query q2 {
  paymentPoint(name_Iexact: "%s", 
                paymentPlan_Uuid: "%s", 
                paymentPoint_Uuid: "%s"
                dateValidFrom: "%s"
                dateValidTo: "%s") {
    edges {
      node {
        id
      }
    }
  }
}
"""

gql_payroll_create = """
mutation createPayroll($name: String!, $paymentCycleId: UUID, $paymentPlanId: UUID!, $paymentPointId: UUID, $paymentMethod: String!, $status: PayrollStatusEnum!, $dateValidFrom: Date, $dateValidTo: Date, $jsonExt: JSONString, $clientMutationId: String) {
  createPayroll(input: {name: $name, paymentCycleId: $paymentCycleId, paymentPlanId: $paymentPlanId, paymentPointId: $paymentPointId, paymentMethod: $paymentMethod, status: $status, dateValidFrom: $dateValidFrom, dateValidTo: $dateValidTo, jsonExt: $jsonExt, clientMutationId : $clientMutationId}) {
    clientMutationId
  }
}
"""

gql_payroll_create_no_json_ext = """
mutation m2 {
  createPayroll (input:{
                name: "%s"
                paymentCycleId: "%s"
                paymentPlanId: "%s" 
                paymentPointId: "%s"
                paymentMethod: "%s"
                status: %s
                dateValidFrom: "%s"
                dateValidTo: "%s"
  }) {
    clientMutationId
  }
}
"""


gql_payroll_delete = """
mutation m2 {
  deletePayroll (input:{
    ids: %s
  }) {
    clientMutationId
  }
}
"""

benefit_consumption_data_test = {
    "photo": "photo-test.jpg",
    "code": "BC123-TEST",
    "date_due": date(2023, 5, 31),
    "receipt": "REC-BC123-TEST",
    "amount": 500.00,
    "type": "Cash",
    "status": "ACCEPTED",
}


benefit_consumption_data_update = {
    "code": "BC123-fixed-fix",
}


gql_benefit_consumption_query = """
query q2 {
  benefitConsumption {
    edges {
      node {
        id
      }
    }
  }
}
"""



import graphene
import graphene_django_optimizer as gql_optimizer
from gettext import gettext as _
from django.contrib.auth.models import AnonymousUser
from django.db.models import Q, Sum

from core.schema import OrderedDjangoFilterConnectionField
from core.services import wait_for_mutation
from core.utils import append_validity_filter
from invoice.gql.gql_types.bill_types import BillGQLType
from invoice.models import Bill
from location.services import get_ancestor_location_filter
from payroll.apps import PayrollConfig
from payroll.gql_mutations import CreatePaymentPointMutation, UpdatePaymentPointMutation, DeletePaymentPointMutation, \
    CreatePayrollMutation, DeletePayrollMutation, ClosePayrollMutation, \
    RejectPayrollMutation, MakePaymentForPayrollMutation, DeleteBenefitConsumptionMutation
from payroll.gql_queries import BenefitConsumptionGQLType, PaymentPointGQLType, \
    PayrollGQLType, PaymentMethodGQLType, \
    PaymentMethodListGQLType, BenefitAttachmentListGQLType, \
    CsvReconciliationUploadGQLType, PayrollBenefitConsumptionGQLType, \
    PaymentGatewayConfigGQLType, BenefitsSummaryGQLType
from payroll.models import PaymentPoint, Payroll, \
    BenefitConsumption, BenefitAttachment, \
    CsvReconciliationUpload, PayrollBenefitConsumption, BenefitConsumptionStatus
from payroll.payments_registry import PaymentMethodStorage
from social_protection.models import BenefitPlan


class Query(graphene.ObjectType):
    payment_point = OrderedDjangoFilterConnectionField(
        PaymentPointGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        client_mutation_id=graphene.String(),
        parent_location=graphene.String(),
        parent_location_level=graphene.Int(),
    )
    payroll = OrderedDjangoFilterConnectionField(
        PayrollGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        client_mutation_id=graphene.String(),
    )

    bill_by_payroll = OrderedDjangoFilterConnectionField(
        BillGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
        payroll_uuid=graphene.UUID(required=True)
    )

    benefit_consumption = OrderedDjangoFilterConnectionField(
        BenefitConsumptionGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        client_mutation_id=graphene.String(),
    )

    payment_methods = graphene.Field(
        PaymentMethodListGQLType,
    )

    payment_gateway_config = graphene.Field(
        PaymentGatewayConfigGQLType,
    )

    benefit_consumption_by_payroll = OrderedDjangoFilterConnectionField(
        BenefitConsumptionGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
        payroll_uuid=graphene.UUID(required=True),
        filterOnlyUnpaid=graphene.Boolean()
    )

    benefit_attachment_by_payroll = OrderedDjangoFilterConnectionField(
        BenefitAttachmentListGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        applyDefaultValidityFilter=graphene.Boolean(),
        client_mutation_id=graphene.String(),
        payroll_uuid=graphene.UUID(required=True)
    )

    csv_reconciliation_upload = OrderedDjangoFilterConnectionField(
        CsvReconciliationUploadGQLType,
        orderBy=graphene.List(of_type=graphene.String),
    )

    payroll_benefit_consumption = OrderedDjangoFilterConnectionField(
        PayrollBenefitConsumptionGQLType,
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        client_mutation_id=graphene.String(),
        benefitPlanName=graphene.String(),
        benefitPlanUuid=graphene.String(),
        paymentCycleUuid=graphene.String(),
    )

    benefits_summary = graphene.Field(
        BenefitsSummaryGQLType,
        individualId=graphene.String(),
        payrollId=graphene.String(),
        benefitPlanUuid=graphene.String(),
        paymentCycleUuid=graphene.String(),
    )

    def resolve_bill_by_payroll(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = [*append_validity_filter(**kwargs), Q(payrollbill__payroll_id=kwargs.get("payroll_uuid"),
                                                        is_deleted=False,
                                                        payrollbill__is_deleted=False,
                                                        payrollbill__payroll__is_deleted=False)]

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        subject_type = kwargs.get("subject_type", None)
        if subject_type:
            filters.append(Q(subject_type__model=subject_type))

        thirdparty_type = kwargs.get("thirdparty_type", None)
        if thirdparty_type:
            filters.append(Q(thirdparty_type__model=thirdparty_type))

        return gql_optimizer.query(Bill.objects.filter(*filters), info)

    def resolve_benefit_consumption_by_payroll(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = [*append_validity_filter(**kwargs),
                   Q(payrollbenefitconsumption__payroll_id=kwargs.get("payroll_uuid"),
                     is_deleted=False,
                     payrollbenefitconsumption__is_deleted=False,
                     payrollbenefitconsumption__payroll__is_deleted=False)]

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        filter_only_unpaid = kwargs.get("filterOnlyUnpaid", None)
        if filter_only_unpaid:
            filters.append(Q(status__in=[BenefitConsumptionStatus.ACCEPTED, BenefitConsumptionStatus.APPROVE_FOR_PAYMENT]))

        return gql_optimizer.query(BenefitConsumption.objects.filter(*filters), info)

    def resolve_benefit_attachment_by_payroll(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = [*append_validity_filter(**kwargs),
                   Q(benefit__payrollbenefitconsumption__payroll_id=kwargs.get("payroll_uuid"),
                     is_deleted=False,
                     benefit__payrollbenefitconsumption__is_deleted=False,
                     benefit__payrollbenefitconsumption__payroll__is_deleted=False)]

        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        return gql_optimizer.query(BenefitAttachment.objects.filter(*filters), info)

    def resolve_payment_point(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payment_point_search_perms)
        filters = []

        client_mutation_id = kwargs.get("client_mutation_id")
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        parent_location = kwargs.get('parent_location')
        if parent_location:
            filters += [get_ancestor_location_filter(parent_location)]

        query = PaymentPoint.objects.filter(*filters)
        return gql_optimizer.query(query, info)

    def resolve_payroll(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id")
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        query = Payroll.objects.filter(*filters)
        return gql_optimizer.query(query, info)

    def resolve_payroll_benefit_consumption(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id")
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        benefit_plan_name = kwargs.get("benefitPlanName")
        if benefit_plan_name:
            benefit_plan_ids = list(BenefitPlan.objects.filter(name__icontains=benefit_plan_name).values_list('id', flat=True))
            filters.append(Q(payroll__payment_plan__benefit_plan_id__in=benefit_plan_ids))

        benefit_plan_uuid = kwargs.get("benefitPlanUuid")
        if benefit_plan_uuid:
            filters.append(Q(payroll__payment_plan__benefit_plan_id=benefit_plan_uuid))

        payment_cycle_uuid = kwargs.get("paymentCycleUuid")
        if payment_cycle_uuid:
            filters.append(Q(payroll__payment_cycle_id=payment_cycle_uuid))

        query = PayrollBenefitConsumption.objects.filter(*filters)
        return gql_optimizer.query(query, info)

    def resolve_benefit_consumption(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_payroll_search_perms)
        filters = append_validity_filter(**kwargs)

        client_mutation_id = kwargs.get("client_mutation_id")
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(Q(mutations__mutation__client_mutation_id=client_mutation_id))

        query = BenefitConsumption.objects.filter(*filters)
        return gql_optimizer.query(query, info)

    def resolve_payment_methods(self, info, **kwargs):
        user = info.context.user
        if type(user) is AnonymousUser or not user.id:
            raise PermissionError("Unauthorized")

        payment_methods = PaymentMethodStorage.get_all_available_payment_methods()
        gql_payment_methods = Query._build_payment_method_options(payment_methods)
        return PaymentMethodListGQLType(gql_payment_methods)

    def resolve_payment_gateway_config(self, info):
        return PaymentGatewayConfigGQLType(
            base_url=PayrollConfig.payment_gateway_base_url,
            api_key=PayrollConfig.payment_gateway_api_key,
            timeout=PayrollConfig.payment_gateway_timeout,
        )

    def resolve_csv_reconciliation_upload(self, info, **kwargs):
        Query._check_permissions(info.context.user, PayrollConfig.gql_csv_reconciliation_search_perms)
        filters = append_validity_filter(**kwargs)

        query = CsvReconciliationUpload.objects.filter(*filters)
        return gql_optimizer.query(query, info)

    def resolve_benefits_summary(self, info, **kwargs):
        Query._check_permissions(info.context.user,
                                 PayrollConfig.gql_payroll_search_perms)
        filters = append_validity_filter(**kwargs)
        individual_id = kwargs.get("individualId", None)
        payroll_id = kwargs.get("payrollId", None)
        benefit_plan_uuid = kwargs.get("benefitPlanUuid", None)
        payment_cycle_uuid = kwargs.get("paymentCycleUuid", None)

        if individual_id:
            filters.append(Q(individual__id=individual_id))

        if payroll_id:
            filters.append(Q(payrollbenefitconsumption__payroll_id=payroll_id))

        if benefit_plan_uuid:
            filters.append(Q(payrollbenefitconsumption__payroll__payment_plan__benefit_plan_id=benefit_plan_uuid))

        if payment_cycle_uuid:
            filters.append(Q(payrollbenefitconsumption__payroll__payment_cycle_id=payment_cycle_uuid))

        amount_received = BenefitConsumption.objects.filter(
            *filters,
            is_deleted=False,
            payrollbenefitconsumption__is_deleted=False,
            status=BenefitConsumptionStatus.RECONCILED
        ).aggregate(total_received=Sum('amount'))['total_received'] or 0

        amount_due = BenefitConsumption.objects.filter(
            *filters,
            is_deleted=False,
            payrollbenefitconsumption__is_deleted=False
        ).exclude(status=BenefitConsumptionStatus.RECONCILED).aggregate(total_due=Sum('amount'))['total_due'] or 0

        return BenefitsSummaryGQLType(
            total_amount_received=amount_received,
            total_amount_due=amount_due,
        )

    @staticmethod
    def _build_payment_method_options(payment_methods):
        gql_payment_methods = []
        for payment_method in payment_methods:
            gql_payment_methods.append(
                PaymentMethodGQLType(
                    name=payment_method['name']
                )
            )
        return gql_payment_methods

    @staticmethod
    def _check_permissions(user, perms):
        if type(user) is AnonymousUser or not user.id or not user.has_perms(perms):
            raise PermissionError(_("Unauthorized"))


class Mutation(graphene.ObjectType):
    create_payment_point = CreatePaymentPointMutation.Field()
    update_payment_point = UpdatePaymentPointMutation.Field()
    delete_payment_point = DeletePaymentPointMutation.Field()

    create_payroll = CreatePayrollMutation.Field()
    delete_payroll = DeletePayrollMutation.Field()
    close_payroll = ClosePayrollMutation.Field()
    reject_payroll = RejectPayrollMutation.Field()
    make_payment_for_payroll = MakePaymentForPayrollMutation.Field()
    delete_benefit_consumption = DeleteBenefitConsumptionMutation.Field()



import json

from graphene import Schema
from graphene.test import Client
from django.test import TestCase

from location.models import Location
from payroll.models import PaymentPoint
from payroll.tests.data import gql_payment_point_query, gql_payment_point_delete, gql_payment_point_update, \
    gql_payment_point_create
from core.test_helpers import LogInHelper
from payroll.schema import Query, Mutation
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase, BaseTestContext


class PaymentPointGQLTestCase(openIMISGraphQLTestCase):


    user = None
    user_unauthorized = None
    gql_client = None
    gql_context = None
    gql_context_unauthorized = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user = LogInHelper().get_or_create_user_api(username='username_authorized', roles=[7])
        cls.user_unauthorized = LogInHelper().get_or_create_user_api(username='username_unauthorized', roles=[1])
        gql_schema = Schema(
            query=Query,
            mutation=Mutation
        )
        cls.gql_client = Client(gql_schema)
        cls.gql_context = BaseTestContext(cls.user)
        cls.gql_context_unauthorized = BaseTestContext(cls.user_unauthorized)
        cls.location = Location.objects.filter(validity_to__isnull=True, type='V').first()

    def test_query(self):
        output = self.gql_client.execute(gql_payment_point_query, context=self.gql_context.get_request())
        result = output.get('data', {}).get('paymentPoint', {})
        self.assertTrue(result)

    def test_query_unauthorized(self):
        output = self.gql_client.execute(gql_payment_point_query, context=self.gql_context_unauthorized.get_request())
        error = next(iter(output.get('errors', [])), {}).get('message', None)
        self.assertTrue(error)

    def test_create(self):
        payload = gql_payment_point_create % (
            json.dumps("Test"),
            self.location.id,
            self.user.id
        )
        output = self.gql_client.execute(payload, context=self.gql_context.get_request())
        self.assertEqual(output.get('errors'), None)
        self.assertTrue(PaymentPoint.objects.filter(
            name="Test", location_id=self.location.id, ppm_id=self.user.id, is_deleted=False).exists())

    def test_create_unauthorized(self):
        payload = gql_payment_point_create % (
            json.dumps("Test"),
            self.location.id,
            self.user.id)
        output = self.gql_client.execute(payload, context=self.gql_context_unauthorized.get_request())
        self.assertFalse(PaymentPoint.objects.filter(
            name="Test", location_id=self.location.id, ppm_id=self.user.id, is_deleted=False).exists())

    def test_update(self):
        payment_point = PaymentPoint(name="Test", location=self.location, ppm=self.user)
        payment_point.save(username=self.user.username)
        payload = gql_payment_point_update % (
            json.dumps(str(payment_point.id)),
            json.dumps("TestUpdated"),
            payment_point.location.id,
            payment_point.ppm.id)
        output = self.gql_client.execute(payload, context=self.gql_context.get_request())
        self.assertEqual(output.get('errors'), None)
        self.assertFalse(PaymentPoint.objects.filter(id=payment_point.id, name="Test", is_deleted=False).exists())
        self.assertTrue(PaymentPoint.objects.filter(id=payment_point.id, name="TestUpdated", is_deleted=False).exists())

    def test_update_unauthorized(self):
        payment_point = PaymentPoint(name="Test", location=self.location, ppm=self.user)
        payment_point.save(username=self.user.username)
        payload = gql_payment_point_update % (
            json.dumps(str(payment_point.id)),
            json.dumps("TestUpdated"),
            payment_point.location.id,
            payment_point.ppm.id)
        output = self.gql_client.execute(payload, context=self.gql_context_unauthorized.get_request())
        self.assertTrue(PaymentPoint.objects.filter(id=payment_point.id, name="Test", is_deleted=False).exists())
        self.assertFalse(
            PaymentPoint.objects.filter(id=payment_point.id, name="TestUpdated", is_deleted=False).exists())

    def test_delete(self):
        payment_point = PaymentPoint(name="Test", location=self.location, ppm=self.user)
        payment_point.save(username=self.user.username)
        payload = gql_payment_point_delete % json.dumps([str(payment_point.id)])
        output = self.gql_client.execute(payload, context=self.gql_context.get_request())
        self.assertEqual(output.get('errors'), None)
        # FIXME self.assertTrue(PaymentPoint.objects.filter(id=payment_point.id, is_deleted=True).exists())

    def test_delete_unauthorized(self):
        payment_point = PaymentPoint(name="Test", location=self.location, ppm=self.user)
        payment_point.save(username=self.user.username)
        payload = gql_payment_point_delete % json.dumps([str(payment_point.id)])
        output = self.gql_client.execute(payload, context=self.gql_context_unauthorized.get_request())
        # FIXME self.assertTrue(PaymentPoint.objects.filter(id=payment_point.id, is_deleted=False).exists())



from django.db import migrations

rolerights = {
    'IMIS Administrator': ['206001', '206002'],
}


def add_rights(role_name, role_model, role_right_model):
    role = role_model.objects.get(name=role_name)
    for right_id in rolerights[role_name]:
        if not role_right_model.objects.filter(validity_to__isnull=True, role=role, right_id=right_id).exists():
            _add_right_for_role(role, right_id, role_right_model)


def _add_right_for_role(role, right_id, role_right_model):
    role_right_model.objects.create(role=role, right_id=right_id, audit_user_id=1)


def on_migration(apps, schema_editor):
    role_model = apps.get_model("core", "role")
    role_right_model = apps.get_model("core", "roleright")
    for role in rolerights:
        add_rights(role, role_model, role_right_model)


class Migration(migrations.Migration):
    dependencies = [
        ('payroll', '0015_merge_20240207_1703'),
    ]

    operations = [
        migrations.RunPython(on_migration),
    ]



from payroll.payment_gateway.payment_gateway_connector import PaymentGatewayConnector


class MockedPaymentGatewayConnector(PaymentGatewayConnector):
    def send_payment(self, invoice_id, amount, **kwargs):
        payload = {"invoiceId": str(invoice_id), "amount": str(amount)}
        response = self.send_request(self.config.endpoint_payment, payload)
        if response:
            response_text = response.text
            expected_message = f"{invoice_id} invoice of {amount} accepted to be paid"
            if response_text == expected_message:
                return True
        return False

    def reconcile(self, invoice_id, amount, **kwargs):
        payload = {"invoiceId": str(invoice_id), "amount": str(amount)}
        response = self.send_request(self.config.endpoint_reconciliation, payload)
        if response:
            response_text = response.text.strip().lower()
            if response_text == "true":
                return True
            elif response_text == "false":
                return False
        return False
